"""
Simple coloring capabilities, without the need of bringing in further dependencies

Example usage:
    from runez import activate_colors, blue

    print(blue("hello"))
"""

import os
import re
import sys

from runez.base import stringified
from runez.convert import shortened, to_int
from runez.system import current_test


# Allows for a clean `import *`
__all__ = [
    "activate_colors", "Color", "color_adjusted_size", "is_coloring", "is_tty", "uncolored",
    "blink", "bold", "dim", "invert", "italic", "strikethrough", "underline",
    "styles",
]


RE_ANSI_ESCAPE = re.compile("\x1b\\[[;\\d]*[A-Za-z]")
BACKEND = None


def is_coloring():
    """
    Returns:
        (bool): True if tty coloring is currently activated
    """
    return BACKEND is not PlainBackend


def is_tty():
    """
    Returns:
        (bool): True if current stdout is a tty
    """
    return (sys.stdout.isatty() or "PYCHARM_HOSTED" in os.environ) and not current_test()


def uncolored(text):
    """
    Args:
        text (str | None): Text to remove ANSI colors from

    Returns:
        (str): Text without any ANSI color escapes
    """
    return RE_ANSI_ESCAPE.sub("", text or "")


def color_adjusted_size(text, size):
    """
    Args:
        text (str): Text to compute color adjusted padding size
        size (int): Desired padding size

    Returns:
        (int): `size`, adjusted to help take into account any color ANSI escapes
    """
    if text:
        size += len(text) - len(uncolored(text))

    return size


class AnsiCode(object):
    """Compute ANSI escape codes to use for a given RGB color"""
    bg_offset = 40
    fg_offset = 30

    @classmethod
    def enhanced(cls, r, g, b, ratio):
        """Enhanced color by 'ratio', allows to brighten/darken colors"""
        f, lm = (min, 255) if ratio > 1 else (max, 0)
        if ratio > 1:
            r = max(8, r)
            g = max(8, g)
            b = max(8, b)

        return f(lm, int(round(r * ratio))), f(lm, int(round(g * ratio))), f(lm, int(round(b * ratio)))

    @classmethod
    def ansi16(cls, offset, r, g, b, brighten):
        """Convert RGB to ANSI 16 color"""
        r = int(round(r / 255.0))
        g = int(round(g / 255.0)) << 1
        b = int(round(b / 255.0)) << 2
        code = (90 if brighten else 30) + (r | g | b)
        return int(code) + offset - cls.fg_offset

    @classmethod
    def ansi256(cls, offset, r, g, b, brighten):
        """Convert RGB to ANSI 256 color"""
        if brighten is not None:
            r, g, b = cls.enhanced(r, g, b, 2 if brighten else 0.7)

        if r == g == b:
            if r < 8:
                code = 16

            elif r > 248:
                code = 231

            else:
                code = round(((r - 8) / 247.0) * 24) + 232

        else:
            r = 36 * round(r / 255.0 * 5.0)
            g = 6 * round(g / 255.0 * 5.0)
            b = round(b / 255.0 * 5.0)
            code = 16 + r + g + b

        return "%s;5;%s" % (8 + offset, int(code))

    @classmethod
    def truecolor(cls, offset, r, g, b, brighten):
        """Convert RGB to ANSI true-color"""
        if brighten is not None:
            r, g, b = cls.enhanced(r, g, b, 1.25 if brighten else 0.8)

        return "%s;2;%s;%s;%s" % (8 + offset, r, g, b)


class ColorFormat(object):
    """Pre-computed color formatters for all supported flavors"""

    def __init__(self, color, fmt):
        self.color = color
        if self.color.rgb is None:
            # `plain` color: return text as-is (no coloring applied)
            self.neutral = "{}"
            self.light = "{}"
            self.dark = "{}"

        else:
            offset = AnsiCode.bg_offset if color.rgb < 0 else AnsiCode.fg_offset
            rgb = abs(color.rgb)
            r = (rgb & 0xff0000) >> 16
            g = (rgb & 0xff00) >> 8
            b = rgb & 0xff

            base_fmt = "\033[{{start}}m{{{{}}}}\033[{end}m".format(end=offset + 9)
            self.neutral = base_fmt.format(start=fmt(offset, r, g, b, None))
            self.light = base_fmt.format(start=fmt(offset, r, g, b, True))
            self.dark = base_fmt.format(start=fmt(offset, r, g, b, False))


class Color(object):
    """Defines a color, with associated tty codes, plus a `name` that can potentially be used by other kinds of backends"""

    def __init__(self, name, rgb):
        """
        Args:
            name (str): Color name (example: blue)
            rgb (int | None): RGB value (example: 0x0000ff)
        """
        self.name = name
        self.rgb = rgb
        self.ansi16 = ColorFormat(self, AnsiCode.ansi16)
        self.ansi256 = ColorFormat(self, AnsiCode.ansi256)
        self.truecolor = ColorFormat(self, AnsiCode.truecolor)

    def __repr__(self):
        return self.name

    def __call__(self, text, shorten=None):
        if shorten:
            text = shortened(text, size=shorten)

        return BACKEND.colored(text, self)


class Style(object):
    def __init__(self, name, start, end):
        self.name = name
        self.start = start
        self.end = end
        self.fmt = "\033[%sm{}\033[%sm" % (start, end)

    def __repr__(self):
        return self.name

    def __call__(self, text, shorten=None):
        if shorten:
            text = shortened(text, size=shorten)

        return BACKEND.styled(text, self)


blink = Style("blink", 5, 25)
bold = Style("bold", 1, 22)
dim = Style("dim", 2, 22)
invert = Style("invert", 7, 27)
italic = Style("italic", 3, 23)
strikethrough = Style("strikethrough", 9, 29)
underline = Style("underline", 4, 24)

styles = [blink, bold, dim, invert, italic, strikethrough, underline]


class PlainBackend(object):
    """Default plain backend, ignoring colors"""
    color_count = 1
    flavor = "neutral"

    @classmethod
    def colored(cls, text, color):
        """
        Args:
            text (str): Text to color
            color (Color): Color to use

        Returns:
            (str): Optionally colored text
        """
        return stringified(text)

    @classmethod
    def styled(cls, text, style):
        """
        Args:
            text (str): Text to color
            style (Style): Style to use

        Returns:
            (str): Optionally styled text
        """
        return stringified(text)


class Ansi16Backend(PlainBackend):
    """ANSI 16-color"""
    color_count = 16
    ansi = "ansi16"

    @classmethod
    def colored(cls, text, color):
        if color is None or text == "":
            return text

        ansi = getattr(color, cls.ansi)
        flavor = getattr(ansi, cls.flavor)
        return flavor.format(text)

    @classmethod
    def styled(cls, text, style):
        if style is None or text == "":
            return text

        return style.fmt.format(text)


class Ansi256Backend(Ansi16Backend):
    """ANSI 256-color"""
    color_count = 256
    ansi = "ansi256"


class TrueColorBackend(Ansi16Backend):
    """ANSI true-color"""
    color_count = 1 << 24
    ansi = "truecolor"


def mentions(value, *terms):
    """Do any of the `terms` mention `value`?"""
    for term in terms:
        if term in value:
            return term


def usable_backends(env):
    """
    Args:
        env (dict): Environemnt variables to inspect

    Returns:
        (list): List of usable backends, give `env`
    """
    result = {PlainBackend}
    if env:
        result.add(Ansi16Backend)
        for env_var in ("COLORTERM", "TERM_PROGRAM", "TERM"):
            value = env.get(env_var)
            if value:
                value = value.lower()
                if mentions(value, "true", "24", "iterm", "hyper"):
                    result.add(TrueColorBackend)

                elif mentions(value, "256", "8", "apple_"):
                    result.add(Ansi256Backend)

    return sorted(result, key=lambda x: -x.color_count)


def detect_flavor(bg):
    """
    Args:
        bg (str | None): Value of env var COLORFGBG

    Returns:
        (str): Flavor to use
    """
    if bg and ";" in bg:
        _, _, bg = bg.partition(";")
        bg = to_int(bg, default=None)
        if bg is not None:
            return "dark" if bg > 6 and bg != 8 else "light"

    return "neutral"


def detect_backend(enable, env):
    """Auto-detect best backend to use"""
    if isinstance(enable, type) and issubclass(enable, PlainBackend):
        return enable

    return usable_backends(enable and env)[0]


def activate_colors(enable, flavor=None, env=os.environ):
    """
    Args:
        enable (bool | PlainBackend.__class__ | None): Set colored output on or off
        flavor (str | None): Flavor to use (neutral, light or dark)
        env (dict): Env vars to use
    """
    global BACKEND
    if enable is None:
        enable = is_tty()
        if flavor is None and enable:
            flavor = detect_flavor(env.get("COLORFGBG"))

    if flavor:
        PlainBackend.flavor = flavor

    BACKEND = detect_backend(enable, env)


activate_colors(None)



from runez.colors.terminal import Color


# Allows for a clean `import *`
__all__ = [
    "black", "blue", "brown", "gray", "green", "orange", "plain", "purple", "red", "teal", "white", "yellow",
    "colors",
]

black = Color("black", 0x000000)
blue = Color("blue", 0x0000ff)
brown = Color("brown", 0x850a0a)
gray = Color("gray", 0xbebebe)
green = Color("green", 0xff00)
orange = Color("orange", 0xef9500)
plain = Color("plain", None)
purple = Color("purple", 0xa020f0)
red = Color("red", 0xff0000)
teal = Color("teal", 0x008080)
white = Color("white", 0xffffff)
yellow = Color("yellow", 0xffff00)

colors = [black, blue, brown, gray, green, orange, purple, red, teal, white, yellow]



#  -*- encoding: utf-8 -*-

import datetime
import re
import time

from runez.base import string_type, stringified, UNSET
from runez.convert import _float_from_text


DEFAULT_TIMEZONE = None
SECONDS_IN_ONE_MINUTE = 60
SECONDS_IN_ONE_HOUR = 60 * SECONDS_IN_ONE_MINUTE
SECONDS_IN_ONE_DAY = 24 * SECONDS_IN_ONE_HOUR
SECONDS_IN_ONE_YEAR = 365.2425 * SECONDS_IN_ONE_DAY

RE_DURATION = re.compile(r"^([0-9]+[ywdhms]\s*)+$")
RE_BASE_NUMBER = r"([-+]?[0-9_]*\.?[0-9_]*([eE][-+]?[0-9_]+)?|[-+]?\.inf|[-+]?\.Inf|[-+]?\.INF|\.nan|\.NaN|\.NAN|0o[0-7]+|0x[0-9a-fA-F]+)"
RE_BASE_DATE = r"(([0-9]{1,4})[-/]([0-9][0-9]?)[-/]([0-9]{1,4})" \
            r"([Tt \t]([0-9][0-9]?):([0-9][0-9]?):([0-9][0-9]?)(\.[0-9]*)?" \
            r"([ \t]*(Z|[A-Z]{3}|[+-][0-9][0-9]?(:([0-9][0-9]?))?))?)?)"

RE_DATE = re.compile("^(%s)$" % "|".join((RE_BASE_NUMBER, RE_BASE_DATE)))


class timezone(datetime.tzinfo):
    """
    There is no handy timezone object available in stdlib.
    We provide this summary implementation in order to mostly support date extraction from text.
    This timezone implementation does not take into account any DST nonsense (do not use this if you care about DST).
    Supported timezone are simply: UTC, and explicit offsets like +01:00
    """

    __singletons = {}  # Cached timezone objects per offset

    def __new__(cls, offset, name=None):
        existing = cls.__singletons.get(offset)
        if existing is None:
            existing = super(timezone, cls).__new__(cls, offset, name=name)
            cls.__singletons[offset] = existing

        return existing

    def __init__(self, offset, name=None):
        if not hasattr(self, "name"):
            self.offset = offset
            if name is None:
                total_seconds = offset.days * SECONDS_IN_ONE_DAY + offset.seconds
                seconds = abs(total_seconds)
                hours = seconds // SECONDS_IN_ONE_HOUR
                seconds -= hours * SECONDS_IN_ONE_HOUR
                minutes = seconds // SECONDS_IN_ONE_MINUTE
                if total_seconds < 0:
                    hours = - hours

                name = "{:+03d}:{:02d}".format(hours, minutes)

            self.name = name

    def __repr__(self):
        return self.name

    def __eq__(self, other):
        if isinstance(other, datetime.tzinfo):
            return self.offset == other.utcoffset(datetime.datetime.now(tz=UTC))

    def utcoffset(self, dt):
        return self.offset

    def tzname(self, dt):
        return self.name

    def dst(self, dt):
        return self.offset


UTC = timezone(datetime.timedelta(0), "UTC")
EPOCH_MS_BREAK = 900000000000
RE_TZ = re.compile(r"([+-]?[0-9][0-9]):?([0-9][0-9])")
DEFAULT_DURATION_SPAN = 2


def date_from_epoch(epoch, in_ms=None):
    """
    Args:
        epoch (int | float): Unix epoch in seconds or milliseconds, utc or local
        in_ms (bool | None): In milliseconds if True, auto-determined if None

    Returns:
        (datetime.date): Corresponding datetime object
    """
    if in_ms or (in_ms is None and epoch > EPOCH_MS_BREAK):
        epoch = epoch / 1000

    return datetime.datetime.utcfromtimestamp(epoch).date()


def datetime_from_epoch(epoch, tz=UNSET, in_ms=None):
    """
    Args:
        epoch (int | float): Unix epoch in seconds or milliseconds, utc or local
        tz (datetime.tzinfo | None): Optional timezone info object, passed through to created datetime
        in_ms (bool | None): In milliseconds if True, auto-determined if None

    Returns:
        (datetime.datetime): Corresponding datetime object
    """
    if tz is UNSET:
        tz = DEFAULT_TIMEZONE

    if in_ms or (in_ms is None and epoch > EPOCH_MS_BREAK):
        epoch = float(epoch) / 1000

    return datetime.datetime.fromtimestamp(epoch, tz=tz)


def elapsed(started, ended=None):
    """
    Args:
        started (datetime.date | datetime.datetime): When operation was started
        ended (datetime.date | datetime.datetime | None): When operation was ended (None means now)

    Returns:
        (float): Elapsed number of seconds
    """
    if not isinstance(started, datetime.datetime):
        tzinfo = None if ended is None else getattr(ended, "tzinfo", None)
        started = datetime.datetime(started.year, started.month, started.day, tzinfo=tzinfo)

    if ended is None:
        ended = datetime.datetime.now(tz=started.tzinfo)

    elif not isinstance(ended, datetime.datetime):
        ended = datetime.datetime(ended.year, ended.month, ended.day, tzinfo=started.tzinfo)

    delta = ended - started
    return delta.total_seconds()


def local_timezone():
    """
    Returns:
        (str): Name of current local timezone
    """
    try:
        return time.tzname[0]

    except (IndexError, TypeError):
        return ""


def represented_duration(seconds, span=UNSET, separator=" "):
    """
    Args:
        seconds (int | float | None): Duration in seconds
        span (int | None): If specified, return `span` most significant parts of the duration, specify <= 0 for short form
                           > 0: N most significant long parts, example: 1 hour 5 seconds
                           None: all parts, example: 1 hour 2 minutes 5 seconds 20 ms
                           0: all parts, short form, example: 1h 2m 5s 20ms
                           < 0: N most significant parts, short form, example: 1h 5s
                           UNSET: use `DEFAULT_DURATION_SPAN` (which can set globally per app, for convenience)
        separator (str): Separator to use between parts

    Returns:
        (str): Human friendly duration representation
    """
    if not isinstance(seconds, (int, float)):
        return stringified(seconds)

    if span is UNSET:
        span = DEFAULT_DURATION_SPAN

    short_form = span is not None and span <= 0
    if span is not None:
        span = abs(span)

    seconds = abs(seconds)
    microseconds = 0 if span and seconds > 10 else int(round((seconds - int(seconds)) * 1000000))
    seconds = int(seconds)

    result = []
    # First, separate seconds and days
    days = seconds // SECONDS_IN_ONE_DAY
    seconds -= days * SECONDS_IN_ONE_DAY

    # Break down days into years, weeks and days
    years = days // 365
    days -= years * 365
    weeks = days // 7
    days -= weeks * 7

    # Break down seconds into hours, minutes and seconds
    hours = seconds // SECONDS_IN_ONE_HOUR
    seconds -= hours * SECONDS_IN_ONE_HOUR
    minutes = seconds // SECONDS_IN_ONE_MINUTE
    seconds -= minutes * SECONDS_IN_ONE_MINUTE

    if years:
        result.append(_duration_span(years, "year", short_form))
    if weeks:
        result.append(_duration_span(weeks, "week", short_form))
    if days:
        result.append(_duration_span(days, "day", short_form))

    if hours:
        result.append(_duration_span(hours, "hour", short_form))
    if minutes:
        result.append(_duration_span(minutes, "minute", short_form))
    if seconds:
        result.append(_duration_span(seconds, "second", short_form))

    if microseconds:
        milliseconds = microseconds // 1000
        microseconds = microseconds % 1000
        if milliseconds:
            result.append(_duration_span(milliseconds, "ms", short_form, immutable=True))
        if microseconds:
            result.append(_duration_span(microseconds, "μs", short_form, immutable=True))

    if not result:
        result.append(_duration_span(seconds, "second", short_form))

    if span:
        result = result[:span]

    return separator.join(result)


def timezone_from_text(value):
    """
    Args:
        value (str | None): Name of timezone, or offset of the form +01:00

    Returns:
        (datetime.tzinfo | None):
    """
    if isinstance(value, timezone):
        return value

    if not value:
        return DEFAULT_TIMEZONE

    if value in ("Z", "UTC"):
        return UTC

    m = RE_TZ.match(value)
    if m:
        hours = int(m.group(1))
        minutes = int(m.group(2))
        return timezone(datetime.timedelta(hours=int(hours), minutes=int(minutes)))

    return DEFAULT_TIMEZONE


def to_date(value):
    """
    Args:
        value: Value to convert to date
        tz (datetime.tzinfo | None): Optional timezone info, used as default if could not be determined from `value`

    Returns:
        (datetime.date | None): Extracted date if possible, otherwise `None`
    """
    if isinstance(value, (int, float)):
        return date_from_epoch(value)

    if isinstance(value, string_type):
        value = _date_from_text(value, date_from_epoch)

    if isinstance(value, datetime.datetime):
        return value.date()

    if isinstance(value, datetime.date):
        return value


def to_datetime(value, tz=UNSET):
    """
    Args:
        value: Value to convert to date
        tz (datetime.tzinfo | None): Optional timezone info, used as default if could not be determined from `value`

    Returns:
        (datetime.datetime | None): Extracted date or datetime if possible, otherwise `None`
    """
    if isinstance(value, (int, float)):
        return datetime_from_epoch(value, tz=tz)

    if isinstance(value, string_type):
        value = _date_from_text(value, datetime_from_epoch, tz=tz)

    if isinstance(value, datetime.datetime):
        return value

    if isinstance(value, datetime.date):
        return datetime.datetime(value.year, value.month, value.day, tzinfo=timezone_from_text(tz))


def to_epoch(date, in_ms=False, tz=UTC):
    """
    Args:
        date (datetime.date | datetime.datetime | None): Date to convert to epoch
        in_ms (bool): If True, return epoch in milliseconds
        tz (datetime.tzinfo | None): Timezone to use for non-datetime `date`-s received

    Returns:
        (int): Epoch in seconds
    """
    if date:
        if not isinstance(date, datetime.datetime):
            date = datetime.datetime(date.year, date.month, date.day, tzinfo=tz)

        ep = (date - datetime.datetime(1970, 1, 1, tzinfo=date.tzinfo)).total_seconds()
        if in_ms:
            return ep * 1000

        return ep


def to_epoch_ms(date, tz=UTC):
    """
    Args:
        date (datetime.date | datetime.datetime): Date to convert to epoch
        tz (datetime.tzinfo | None): Timezone to use for non-datetime `date`-s received

    Returns:
        (int): Epoch in seconds
    """
    return to_epoch(date, in_ms=True, tz=tz)


def to_seconds(duration):
    """
    Args:
        duration (str | int | datetime.timedelta | None): Text representing duration, like 30m or 1h or 1h30m
            Accepted input if of the form <number><unit>, N times
            Possible units are: w: weeks, d: days, h: hours, m: minutes, s: seconds

    Returns:
        (int | None): Duration in seconds
    """
    if isinstance(duration, (int, float)):
        return duration

    if isinstance(duration, datetime.timedelta):
        return duration.total_seconds()

    if not isinstance(duration, string_type):
        return None

    duration = duration.strip()
    if not duration:
        return 0

    m = RE_DURATION.match(duration)
    if not m:
        dt = to_datetime(duration)
        if dt is not None:
            return elapsed(dt)

        return None

    v = m.group(1)
    seconds = to_seconds(duration.replace(v, ""))

    # v = v.strip()
    if v.endswith("w"):
        seconds += int(v[:-1], 0) * SECONDS_IN_ONE_DAY * 7

    elif v.endswith("d"):
        seconds += int(v[:-1], 0) * SECONDS_IN_ONE_DAY

    elif v.endswith("h"):
        seconds += int(v[:-1]) * SECONDS_IN_ONE_HOUR

    elif v.endswith("m"):
        seconds += int(v[:-1]) * SECONDS_IN_ONE_MINUTE

    elif v.endswith("y"):
        seconds += int(v[:-1]) * SECONDS_IN_ONE_YEAR

    else:
        seconds += int(v[:-1])

    return seconds


def _date_from_components(components, tz=UNSET):
    """
    Args:
        components (tuple): Components from regex
        tz (datetime.tzinfo | None): Optional timezone info, used as default if could not be determined from `components`

    Returns:
        (datetime.date | datetime.datetime | None)
    """
    y, m, d, _, hh = components[:5]

    try:
        y = int(y)
        m = int(m)
        d = int(d)
        if d > 100 and y < 100:
            # Best effort: allow for european-style notation month/day/year
            m, d, y = y, m, d

        if hh is None:
            return datetime.date(y, m, d)

        mm, ss, sf, _, ctz = components[5:10]
        hh = int(hh)
        mm = int(mm)
        ss = int(ss)
        sf = int(round(float(sf or 0) * 1000000))
        return datetime.datetime(y, m, d, hh, mm, ss, sf, timezone_from_text(ctz or tz))

    except ValueError:
        # Funky date style, ignore
        return None


def _date_from_text(text, epocher, **kwargs):
    """
    Args:
        text (str): Value to turn into date or datetime
        epocher (callable): Function to use to transform int to date or datetime

    Returns:
        (datetime.date | datetime.datetime | None): Extracted date, if possible
    """
    match = RE_DATE.match(text)
    if match is None:
        m = RE_DURATION.match(text)
        if m:
            tz = kwargs.get("tz", UTC)
            offset = to_seconds(text)
            now = datetime.datetime.now(tz=tz)
            return to_datetime(to_epoch(now) - offset, tz=tz)

        return None

    # _, number, _, _, y, m, d, _, hh, mm, ss, sf, _, tz, _, _ = match.groups()
    components = match.groups()
    if components[1]:
        return epocher(_float_from_text(components[1], lenient=True), **kwargs)

    return _date_from_components(components[4:14], **kwargs)


def _duration_span(count, name, short_form, immutable=False):
    if short_form:
        if not immutable:
            name = name[0]

    else:
        name = " %s%s" % (name, "" if immutable or count == 1 else "s")

    return "%s%s" % (count, name)



"""
Convenience methods for file/folder operations
"""

import logging
import os

from runez.convert import resolved_path, short
from runez.system import abort, is_dryrun


LOG = logging.getLogger(__name__)


def basename(path, extension_marker="."):
    """Base name of given `path`, ignoring extension if `extension_marker` is provided

    Args:
        path (str | None): Path to consider
        extension_marker (str | None): Also trim file extension, if marker provided

    Returns:
        (str): Basename part of path, without extension (if 'extension_marker' provided)
    """
    result = os.path.basename(path or "")
    if extension_marker:
        pre, _, post = result.rpartition(extension_marker)
        return pre or post

    return result


def ensure_folder(path, folder=False, fatal=True, logger=LOG.debug, dryrun=None):
    """Ensure folder exists

    Args:
        path (str | None): Path to file or folder
        folder (bool): If True, 'path' refers to a folder (file assumed otherwise)
        fatal (bool | None): Abort execution on failure if True
        logger (callable | None): Logger to use
        dryrun (bool | None): If specified, override global is_dryrun()

    Returns:
         (int): 1 if effectively done, 0 if no-op, -1 on failure
    """
    if not path:
        return 0

    if folder:
        folder = resolved_path(path)

    else:
        folder = parent_folder(path)

    if os.path.isdir(folder):
        if not os.access(folder, os.W_OK):
            return abort("Folder %s is not writable", folder, fatal=(fatal, -1), logger=logger)
        return 0

    if dryrun is None:
        dryrun = is_dryrun()

    if dryrun:
        LOG.debug("Would create %s", short(folder))
        return 1

    try:
        os.makedirs(folder)
        if logger:
            logger("Created folder %s", short(folder))

        return 1

    except Exception as e:
        return abort("Can't create folder %s: %s", short(folder), e, fatal=(fatal, -1), logger=logger)


def parent_folder(path, base=None):
    """Parent folder of `path`, relative to `base`

    Args:
        path (str | None): Path to file or folder
        base (str | None): Base folder to use for relative paths (default: current working dir)

    Returns:
        (str): Absolute path of parent folder
    """
    return path and os.path.dirname(resolved_path(path, base=base))



import io
import logging
import os
import shutil
import tempfile

from runez.base import decode
from runez.convert import Anchored, resolved_path, short, SYMBOLIC_TMP, to_int
from runez.path import ensure_folder, parent_folder
from runez.system import abort, is_dryrun, set_dryrun

LOG = logging.getLogger(__name__)
TEXT_THRESHOLD_SIZE = 1048576  # Max size in bytes to consider a file a "text file"


def copy(source, destination, ignore=None, adapter=None, fatal=True, logger=LOG.debug):
    """Copy source -> destination

    Args:
        source (str | None): Source file or folder
        destination (str | None): Destination file or folder
        ignore (callable | list | str | None): Names to be ignored
        adapter (callable | None): Optional function to call on 'source' before copy
        fatal (bool | None): Abort execution on failure if True
        logger (callable | None): Logger to use

    Returns:
        (int): 1 if effectively done, 0 if no-op, -1 on failure
    """
    return _file_op(source, destination, _copy, adapter, fatal, logger, ignore=ignore)


def delete(path, fatal=True, logger=LOG.debug):
    """
    Args:
        path (str | None): Path to file or folder to delete
        fatal (bool | None): Abort execution on failure if True
        logger (callable | None): Logger to use

    Returns:
        (int): 1 if effectively done, 0 if no-op, -1 on failure
    """
    path = resolved_path(path)
    islink = path and os.path.islink(path)
    if not islink and (not path or not os.path.exists(path)):
        return 0

    if is_dryrun():
        LOG.debug("Would delete %s", short(path))
        return 1

    if logger:
        logger("Deleting %s", short(path))

    try:
        if islink or os.path.isfile(path):
            os.unlink(path)

        else:
            shutil.rmtree(path)

        return 1

    except Exception as e:
        return abort("Can't delete %s: %s", short(path), e, fatal=(fatal, -1))


def first_line(path, default=None):
    """First line of file with `path`, if any

    Args:
        path (str | None): Path to file
        default (str | None): Default to return if file could not be read

    Returns:
        (str | None): First line of file, if any
    """
    try:
        with io.open(resolved_path(path), errors="ignore") as fh:
            return fh.readline().strip()

    except (IOError, TypeError, ValueError):
        return default


def ini_to_dict(data, keep_empty=False, default=None):
    """Contents of an INI-style config file as a dict of dicts: section -> key -> value

    Args:
        data (str | file | list | None): Path to file, or file object, or lines to parse
        keep_empty (bool): If True, keep definitions with empty values
        default (dict | None): Object to return if conf couldn't be read

    Returns:
        (dict): Dict of section -> key -> value
    """
    if not data:
        return default

    lines = readlines(data)
    if lines is None:
        return default

    result = {}
    section_key = None
    section = None
    for line in lines:
        line = decode(line).strip()
        if "#" in line:
            i = line.index("#")
            line = line[:i].strip()

        if not line:
            continue

        if line.startswith("[") and line.endswith("]"):
            section_key = line.strip("[]").strip()
            section = result.get(section_key)
            continue

        if "=" not in line:
            continue

        if section is None:
            section = result[section_key] = {}

        key, _, value = line.partition("=")
        key = key.strip()
        value = value.strip()
        if keep_empty or (key and value):
            section[key] = value

    if not keep_empty:
        result = dict((k, v) for k, v in result.items() if k and v)

    return result


def readlines(data, max_size=TEXT_THRESHOLD_SIZE, default=None):
    """Tentatively read lines from `data`, if not possible, simply return `default`

    Args:
        data (str | file | list | None): Path to file, or file object to return lines from
        max_size (int | None): Return contents only for files smaller than 'max_size' bytes
        default (list | None): Object to return if lines couldn't be read

    Returns:
        (list | None): Lines from file contents
    """
    if not data:
        return default

    if isinstance(data, list):
        return data

    if hasattr(data, "readlines"):
        return data.readlines()

    path = resolved_path(data)
    if not os.path.isfile(path) or (max_size and os.path.getsize(path) > max_size):
        # Intended for small text files, pretend no contents for binaries
        return default

    try:
        with io.open(path) as fh:
            return fh.readlines()

    except Exception:
        return default


def move(source, destination, adapter=None, fatal=True, logger=LOG.debug):
    """Move `source` -> `destination`

    Args:
        source (str | None): Source file or folder
        destination (str | None): Destination file or folder
        adapter (callable): Optional function to call on 'source' before copy
        fatal (bool | None): Abort execution on failure if True
        logger (callable | None): Logger to use

    Returns:
        (int): 1 if effectively done, 0 if no-op, -1 on failure
    """
    return _file_op(source, destination, _move, adapter, fatal, logger)


def symlink(source, destination, adapter=None, must_exist=True, fatal=True, logger=LOG.debug):
    """Symlink `source` <- `destination`

    Args:
        source (str | None): Source file or folder
        destination (str | None): Destination file or folder
        adapter (callable): Optional function to call on 'source' before copy
        must_exist (bool): If True, verify that source does indeed exist
        fatal (bool | None): Abort execution on failure if True
        logger (callable | None): Logger to use

    Returns:
        (int): 1 if effectively done, 0 if no-op, -1 on failure
    """
    return _file_op(source, destination, _symlink, adapter, fatal, logger, must_exist=must_exist)


class TempFolder(object):
    """Context manager for obtaining a temp folder"""

    def __init__(self, anchor=True, dryrun=None, follow=True):
        """
        Args:
            anchor (bool): If True, short-ify paths relative to used temp folder
            dryrun (bool): Override dryrun (if provided)
            follow (bool): If True, change working dir to temp folder (and restore)
        """
        self.anchor = anchor
        self.dryrun = dryrun
        self.follow = follow
        self.old_cwd = None
        self.tmp_folder = None

    def __enter__(self):
        if self.dryrun is not None:
            self.dryrun = set_dryrun(self.dryrun)

        if not is_dryrun():
            # Use realpath() to properly resolve for example symlinks on OSX temp paths
            self.tmp_folder = os.path.realpath(tempfile.mkdtemp())
            if self.follow:
                self.old_cwd = os.getcwd()
                os.chdir(self.tmp_folder)

        tmp = self.tmp_folder or SYMBOLIC_TMP
        if self.anchor:
            Anchored.add(tmp)

        return tmp

    def __exit__(self, *_):
        if self.anchor:
            Anchored.pop(self.tmp_folder or SYMBOLIC_TMP)

        if self.old_cwd:
            os.chdir(self.old_cwd)

        if self.tmp_folder:
            shutil.rmtree(self.tmp_folder)

        if self.dryrun is not None:
            set_dryrun(self.dryrun)


def terminal_width(default=None):
    """Get the width (number of columns) of the terminal window.

    Args:
        default: Default to use if terminal width could not be determined

    Returns:
        (int): Determined terminal width, if possible
    """
    for func in (_tw_shutil, _tw_env):
        columns = func()
        if columns is not None:
            return columns

    return to_int(default)


def touch(path, fatal=True, logger=None):
    """Touch file with `path`

    Args:
        path (str | None): Path to file to touch
        fatal (bool | None): Abort execution on failure if True
        logger (callable | None): Logger to use

    Returns:
        (int): 1 if effectively done, 0 if no-op, -1 on failure
    """
    return write(path, None, fatal=fatal, logger=logger)


def write(path, contents, fatal=True, logger=None):
    """Write `contents` to file with `path`

    Args:
        path (str | None): Path to file
        contents (str | None): Contents to write (only touch file if None)
        fatal (bool | None): Abort execution on failure if True
        logger (callable | None): Logger to use

    Returns:
        (int): 1 if effectively done, 0 if no-op, -1 on failure
    """
    if not path:
        return 0

    path = resolved_path(path)
    if is_dryrun():
        action = "write %s bytes to" % len(contents) if contents else "touch"
        LOG.debug("Would %s %s", action, short(path))
        return 1

    ensure_folder(path, fatal=fatal, logger=logger)
    if logger and contents:
        logger("Writing %s bytes to %s", len(contents), short(path))

    try:
        with io.open(path, "wt") as fh:
            if contents is None:
                os.utime(path, None)

            else:
                fh.write(decode(contents))

        return 1

    except Exception as e:
        return abort("Can't write to %s: %s", short(path), e, fatal=(fatal, -1))


def _copy(source, destination, ignore=None):
    """Effective copy"""
    if os.path.isdir(source):
        if os.path.isdir(destination):
            for fname in os.listdir(source):
                _copy(os.path.join(source, fname), os.path.join(destination, fname), ignore=ignore)
        else:
            if os.path.isfile(destination) or os.path.islink(destination):
                os.unlink(destination)
            shutil.copytree(source, destination, symlinks=True, ignore=ignore)
    else:
        shutil.copy(source, destination)

    shutil.copystat(source, destination)  # Make sure last modification time is preserved


def _move(source, destination):
    """Effective move"""
    shutil.move(source, destination)


def _symlink(source, destination):
    """Effective symlink"""
    os.symlink(source, destination)


def _file_op(source, destination, func, adapter, fatal, logger, must_exist=True, ignore=None):
    """Call func(source, destination)

    Args:
        source (str | None): Source file or folder
        destination (str | None): Destination file or folder
        func (callable): Implementation function
        adapter (callable | None): Optional function to call on 'source' before copy
        fatal (bool | None): Abort execution on failure if True
        logger (callable | None): Logger to use
        must_exist (bool): If True, verify that source does indeed exist
        ignore (callable | list | str | None): Names to be ignored

    Returns:
        (int): 1 if effectively done, 0 if no-op, -1 on failure
    """
    if not source or not destination or source == destination:
        return 0

    action = func.__name__[1:]
    indicator = "<-" if action == "symlink" else "->"
    psource = parent_folder(source)
    pdest = resolved_path(destination)
    if psource != pdest and psource.startswith(pdest):
        return abort(
            "Can't %s %s %s %s: source contained in destination", action, short(source), indicator, short(destination), fatal=(fatal, -1)
        )

    if is_dryrun():
        LOG.debug("Would %s %s %s %s", action, short(source), indicator, short(destination))
        return 1

    if must_exist and not os.path.exists(source):
        return abort("%s does not exist, can't %s to %s", short(source), action.title(), short(destination), fatal=(fatal, -1))

    try:
        # Ensure parent folder exists
        ensure_folder(destination, fatal=fatal, logger=None)

        if logger:
            note = adapter(source, destination, fatal=fatal, logger=logger) if adapter else ""
            if logger:
                logger("%s %s %s %s%s", action.title(), short(source), indicator, short(destination), note)

        if ignore is not None:
            if callable(ignore):
                func(source, destination, ignore=ignore)

            else:
                func(source, destination, ignore=lambda *_: ignore)

        else:
            func(source, destination)

        return 1

    except Exception as e:
        return abort("Can't %s %s %s %s: %s", action, short(source), indicator, short(destination), e, fatal=(fatal, -1))


def _tw_shutil():
    try:
        return shutil.get_terminal_size().columns

    except Exception:
        return None


def _tw_env():
    return to_int(os.environ.get("COLUMNS"))



"""
Convenience logging setup
"""

import logging
import os
import signal
import sys
import threading
from functools import partial
from logging.handlers import RotatingFileHandler, TimedRotatingFileHandler

try:
    # faulthandler is only available in python 3.3+
    import faulthandler
    from typing import List, Optional

except ImportError:
    faulthandler = None

from runez.base import Slotted, ThreadGlobalContext, UNSET
from runez.convert import flattened, formatted, represented_args, SANITIZED, to_bytesize, to_int, UNIQUE
from runez.date import local_timezone
from runez.path import basename as get_basename, ensure_folder, parent_folder
from runez.program import dev_folder, program_path
from runez.system import is_dryrun, set_dryrun, WINDOWS


LOG = logging.getLogger(__name__)
ORIGINAL_CF = logging.currentframe


class LoggingSnapshot(Slotted):
    """
    Take a snapshot of parts we're modifying in the 'logging' module, in order to be able to restore it as it was
    """

    __slots__ = ["_srcfile", "critical", "fatal", "error", "exception", "warning", "info", "debug"]

    def _seed(self):
        """Seed initial fields"""
        for name in self.__slots__:
            setattr(self, name, getattr(logging, name))

    def restore(self):
        for name in self.__slots__:
            setattr(logging, name, getattr(self, name))


class LogSpec(Slotted):
    """
    Settings to use, you can safely customize these (before calling runez.log.setup), for example:
        runez.log.setup(console_stream=sys.stdout)
    or
        runez.log.spec.set(console_stream=sys.stdout)
        runez.log.setup()
    or
        runez.log.spec.console_stream = sys.stdout
        runez.log.setup()
    """

    # See setup()'s docstring for meaning of each field
    __slots__ = [
        "appname",
        "basename",
        "console_format",
        "console_level",
        "console_stream",
        "context_format",
        "dev",
        "file_format",
        "file_level",
        "file_location",
        "locations",
        "rotate",
        "rotate_count",
        "timezone",
        "tmp",
    ]

    @property
    def argv(self):
        """str: Command line invocation, represented to show as greeting"""
        return represented_args(sys.argv)

    @property
    def pid(self):
        """str: Current process id represented to show as greeting"""
        return os.getpid()

    def usable_location(self):
        """
        Returns:
            str | None: First available usable location
        """
        if not self.should_log_to_file:
            return None

        if self.file_location is not None:
            # Custom location typically provided via --config CLI flag
            return self._auto_complete_filename(self.file_location)

        if self.locations:
            for location in self.locations:
                path = self._auto_complete_filename(location)
                if path:
                    return path

    @property
    def should_log_to_file(self):
        """
        Returns:
            bool: As per the spec, should we be logging to a file?
        """
        if not self.file_format:
            return False

        if self.file_location is not None:
            return bool(self.file_location)

        return bool(self.locations)

    def _auto_complete_filename(self, location):
        """
        Args:
            location (str | unicode | None): Location to auto-complete with {basename}, if it points to a folder

        Returns:
            str | None: {location}/{basename}
        """
        path = formatted(location, self, os.environ)
        if path:
            if os.path.isdir(path):
                filename = formatted(self.basename, self)
                if not filename:
                    return None

                path = os.path.join(path, filename)

            folder = parent_folder(parent_folder(path))
            if folder and os.path.exists(folder):
                # Try to auto-create log folder only if "grand parent" folder exists
                if ensure_folder(path, fatal=False, logger=LOG.debug, dryrun=False) >= 0:
                    return path


class _ContextFilter(logging.Filter):
    """
    Optional logging filter allowing to inject key/value pairs to every log record.

    In order to activate this:
    - Mention %(context)s in your log format
    - Add key/value pairs via runez.log.context.add_global(), runez.log.context.add_threadlocal()
    """

    def __init__(self, context, name=""):
        """
        Args:
            context (ThreadGlobalContext): Associated context
            name (str | unicode): Passed through to parent
        """
        super(_ContextFilter, self).__init__(name=name)
        self.context = context

    def filter(self, record):
        """Determines if the record should be logged and injects context info into the record. Always returns True"""
        fmt = LogManager.spec.context_format
        if fmt:
            data = self.context.to_dict()
            if data:
                record.context = fmt % ",".join("%s=%s" % (key, val) for key, val in sorted(data.items()) if key and val)

            else:
                record.context = ""

        return True


def default_log_locations():
    if WINDOWS:  # pragma: no cover
        return [os.path.join("{dev}", "log", "{basename}")]

    return ["{dev}/log/{basename}", "/logs/{appname}/{basename}", "/var/log/{basename}"]


class LogManager(object):
    """
    Global logging context managed by runez.
    There's only one, as multiple contexts would not be useful (logging setup is a global thing)
    """

    # Defaults used to initialize LogSpec instances
    # Use runez.log.override_spec() to change these defaults (do not change directly)
    _default_spec = LogSpec(
        appname=None,
        basename="{appname}.log",
        console_format="%(asctime)s %(levelname)s %(message)s",
        console_level=logging.WARNING,
        console_stream=sys.stderr,
        context_format="[[%s]] ",
        dev=None,
        file_format="%(asctime)s %(timezone)s [%(threadName)s] %(context)s%(levelname)s - %(message)s",
        file_level=logging.DEBUG,
        file_location=None,
        locations=default_log_locations(),
        rotate=None,
        rotate_count=10,
        timezone=local_timezone(),
        tmp=None,
    )

    # Spec defines how logs should be setup()
    # Best way to provide your spec is via: runez.log.setup(), for example:
    #   runez.log.setup(rotate="size:50m")
    spec = LogSpec(_default_spec)

    # Thread-local / global context
    context = ThreadGlobalContext(_ContextFilter)

    # Below fields should be read-only for outside users, do not modify these
    debug = None
    console_handler = None  # type: Optional[logging.StreamHandler]
    file_handler = None  # type: Optional[logging.FileHandler] # File we're currently logging to (if any)
    handlers = None  # type: Optional[List[logging.Handler]]
    used_formats = None  # type: Optional[str]
    faulthandler_signum = None  # type: Optional[int]

    _lock = threading.RLock()
    _logging_snapshot = LoggingSnapshot()
    _shortcuts_fixed = False

    @classmethod
    def set_debug(cls, debug):
        """Useful only as simple callback function, use rune.log.setup() for regular usage"""
        if debug is None and is_dryrun():
            debug = True

        cls.set_debug_dryrun(debug=debug)

    @classmethod
    def set_dryrun(cls, dryrun):
        """Useful only as simple callback function, use rune.log.setup() for regular usage"""
        cls.set_debug_dryrun(dryrun=dryrun)

    @classmethod
    def set_file_location(cls, file_location):
        """Useful only as simple callback function, use rune.log.setup() for regular usage"""
        LogManager.override_spec(file_location=file_location)

    @classmethod
    def set_debug_dryrun(cls, debug=UNSET, dryrun=UNSET):
        """
        Args:
            debug (bool): Enable debug level logging (overrides other specified levels)
            dryrun (bool): Enable dryrun
        """
        if dryrun is not UNSET:
            if dryrun and (debug is None or debug is UNSET):
                # Automatically turn debug on (if not explicitly specified) with dryrun,
                # as many of the "Would ..." messages are at debug level
                debug = True

            set_dryrun(dryrun)

        if debug is not UNSET:
            cls.debug = debug

    @classmethod
    def setup(
            cls,
            debug=UNSET,
            dryrun=UNSET,
            level=UNSET,
            clean_handlers=UNSET,
            greetings=UNSET,
            appname=UNSET,
            basename=UNSET,
            console_format=UNSET,
            console_level=UNSET,
            console_stream=UNSET,
            context_format=UNSET,
            dev=UNSET,
            file_format=UNSET,
            file_level=UNSET,
            file_location=UNSET,
            locations=UNSET,
            rotate=UNSET,
            rotate_count=UNSET,
            timezone=UNSET,
            tmp=UNSET,
    ):
        """
        Args:
            debug (bool): Enable debug level logging (overrides other specified levels)
            dryrun (bool): Enable dryrun
            level (int | None): Shortcut to set both `console_level` and `file_level` at once
            clean_handlers (bool): Remove any existing logging.root.handlers
            greetings (str | unicode | list[str | unicode] | None): Optional greetings message(s) to log
            appname (str | unicode | None): Program's base name, not used directly, just as reference for default 'basename'
            basename (str | unicode | None): Base name of target log file, not used directly, just as reference for default 'locations'
            console_format (str | unicode | None): Format to use for console log, use None to deactivate
            console_level (int | None): Level to use for console logging
            console_stream (TextIOWrapper | None): Stream to use for console log (eg: sys.stderr), use None to deactivate
            context_format (str | unicode | None): Format to use for contextual log, use None to deactivate
            dev (str | unicode | None): Custom folder to use when running from a development venv (auto-determined if None)
            file_format (str | unicode | None): Format to use for file log, use None to deactivate
            file_level (int | None): Level to use for file logging
            file_location (str | unicode | None): Desired custom file location (overrides {locations} search, handy as a --log cli flag)
            locations (list[str | unicode]|None): List of candidate folders for file logging (None: deactivate file logging)
            rotate (str | unicode | None): How to rotate log file (None: no rotation, "time:1d" time-based, "size:50m" size-based)
            rotate_count (int): How many rotations to keep
            timezone (str | unicode | None): Time zone, use None to deactivate time zone logging
            tmp (str | unicode | None): Optional temp folder to use (auto determined)
        """
        with cls._lock:
            cls.set_debug_dryrun(debug=debug, dryrun=dryrun)
            cls.spec.set(
                appname=appname,
                basename=basename,
                console_format=console_format,
                console_level=console_level or level,
                console_stream=console_stream,
                context_format=context_format,
                dev=dev,
                file_format=file_format,
                file_level=file_level or level,
                file_location=file_location,
                locations=locations,
                rotate=rotate,
                rotate_count=rotate_count,
                timezone=timezone,
                tmp=tmp,
            )

            cls._auto_fill_defaults()

            if cls.debug:
                cls.spec.console_level = logging.DEBUG
                cls.spec.file_level = logging.DEBUG

            elif level:
                cls.spec.console_level = level
                cls.spec.file_level = level

            root_level = min(flattened([cls.spec.console_level, cls.spec.file_level], split=SANITIZED))
            if root_level and root_level != logging.root.level:
                logging.root.setLevel(root_level)

            if cls.handlers is None:
                cls.handlers = []

            cls._setup_handler("console")
            cls._setup_handler("file")
            cls._update_used_formats()
            cls._fix_logging_shortcuts()

            if clean_handlers:
                cls.clean_handlers()

            cls.greet(greetings)

    @classmethod
    def greet(cls, greetings, logger=LOG.debug):
        """
        Args:
            greetings (str | unicode | list[str | unicode] | None): Greetings message(s) to log
            logger (callable | None): Logger to use
        """
        if greetings and logger:
            for msg in flattened(greetings, split=SANITIZED):
                message = cls.formatted_greeting(msg)
                if message:
                    logger(message)

    @classmethod
    def clean_handlers(cls):
        """Remove all non-runez logging handlers"""
        for h in logging.root.handlers:
            if h is not cls.console_handler and h is not cls.file_handler:
                logging.root.removeHandler(h)

    @classmethod
    def reset(cls):
        """Reset logging as it was before setup(), no need to call this outside of testing, or some very special cases"""
        cls._disable_faulthandler()
        if cls.handlers is not None:
            for handler in cls.handlers:
                logging.root.removeHandler(handler)

            cls.handlers = None

        cls._logging_snapshot.restore()
        cls.context.reset()
        cls.spec = LogSpec(cls._default_spec)
        cls.debug = None
        cls.console_handler = None
        cls.file_handler = None
        cls.used_formats = None

    @classmethod
    def formatted_greeting(cls, greeting):
        if greeting:
            if cls.file_handler and cls.file_handler.baseFilename:
                location = cls.file_handler.baseFilename

            elif not cls.spec.should_log_to_file:
                location = "file log disabled"

            elif cls.spec.file_location:
                location = "{file_location} is not usable"

            else:
                location = "no usable locations from {locations}"

            return formatted(greeting, cls.spec, location=location, strict=False)

    @classmethod
    def silence(cls, *modules, **kwargs):
        """
        Args:
            *modules: Modules, or names of modules to silence (by setting their log level to WARNING or above)
            **kwargs: Pass as kwargs due to python 2.7, would be level=logging.WARNING otherwise
        """
        level = kwargs.pop("level", logging.WARNING)
        for mod in modules:
            name = mod.__name__ if hasattr(mod, "__name__") else mod
            logging.getLogger(name).setLevel(level)

    @classmethod
    def is_using_format(cls, markers, used_formats=None):
        """
        Args:
            markers (str | unicode): Space separated list of markers to look for
            used_formats (str | unicode): Formats to consider (default: cls.used_formats)

        Returns:
            (bool): True if any one of the 'markers' is seen in 'used_formats'
        """
        if used_formats is None:
            used_formats = cls.used_formats

        if not markers or not used_formats:
            return False

        return any(marker in used_formats for marker in flattened(markers, split=(" ", UNIQUE)))

    @classmethod
    def enable_faulthandler(cls, signum=getattr(signal, "SIGUSR1", None)):
        """Enable dumping thread stack traces when specified signals are received, similar to java's handling of SIGQUIT

        Note: this must be called from the surviving process in case of daemonization.

        Args:
            signum (int | None): Signal number to register for full thread stack dump (use None to disable)
        """
        with cls._lock:
            if not signum:
                cls._disable_faulthandler()
                return

            if not cls.file_handler or faulthandler is None:
                return

            cls.faulthandler_signum = signum
            dump_file = cls.file_handler.stream
            faulthandler.enable(file=dump_file, all_threads=True)
            faulthandler.register(signum, file=dump_file, all_threads=True, chain=False)

    @classmethod
    def override_spec(cls, **kwargs):
        """OVerride 'spec' and '_default_spec' with given values"""
        cls._default_spec.set(**kwargs)
        cls.spec.set(**kwargs)

    @classmethod
    def _update_used_formats(cls):
        cls.used_formats = None
        for handler in (cls.console_handler, cls.file_handler):
            if handler and handler.formatter and handler.formatter._fmt:
                cls.used_formats = "%s %s" % (cls.used_formats or "", handler.formatter._fmt)
                cls.used_formats = cls.used_formats.strip()

    @classmethod
    def _setup_handler(cls, name):
        fmt = _canonical_format(getattr(cls.spec, "%s_format" % name))
        level = getattr(cls.spec, "%s_level" % name)
        if name == "console":
            target = cls.spec.console_stream

        else:
            target = cls.spec.usable_location()

        existing = getattr(cls, "%s_handler" % name)
        if cls._is_equivalent_handler(existing, target, fmt, level):
            return

        if existing is not None:
            cls.handlers.remove(existing)
            logging.root.removeHandler(existing)

        if not target:
            return

        if name == "console":
            new_handler = logging.StreamHandler(target)

        else:
            new_handler = _get_file_handler(target, cls.spec.rotate, cls.spec.rotate_count)

        if fmt:
            new_handler.setFormatter(logging.Formatter(fmt))

        if level:
            new_handler.setLevel(level)

        setattr(cls, "%s_handler" % name, new_handler)
        logging.root.addHandler(new_handler)
        cls.handlers.append(new_handler)

    @classmethod
    def _is_equivalent_handler(cls, existing, target, fmt, level):
        """
        Args:
            existing (logging.Handler): Existing handler to examine
            target: Target for candidate new handler
            fmt: Format for candidate new handler
            level: Level for candidate new handler

        Returns:
            (bool): True if `existing` handler is equivalent to candidate new handler
        """
        if existing is None:
            return False

        if not existing.formatter or fmt != existing.formatter._fmt:
            return False

        if existing.level != level:
            return False

        if isinstance(existing, logging.FileHandler):
            return existing.baseFilename == target

        if isinstance(existing, logging.StreamHandler):
            return existing.stream == target

    @classmethod
    def _auto_fill_defaults(cls):
        """Late auto-filled missing defaults (caller's value kept if provided)"""
        if not cls.spec.appname:
            cls.spec.appname = get_basename(program_path())

        if not cls.spec.dev:
            cls.spec.dev = dev_folder()

    @classmethod
    def _disable_faulthandler(cls):
        if faulthandler and cls.faulthandler_signum:
            faulthandler.unregister(cls.faulthandler_signum)
            faulthandler.disable()
            cls.faulthandler_signum = None

    @classmethod
    def _fix_logging_shortcuts(cls):
        """
        Fix standard logging shortcuts to correctly report logging module.

        This is only useful if you:
        - actually use %(name) and care about it being correct
        - you would still like to use the logging.info() etc shortcuts

        So basically you'd like to write this:
            import logging
            logging.info("hello")

        Instead of this:
            import logging
            LOG = logging.getLogger(__name__)
            LOG.info("hello")
        """
        if cls.is_using_format("%(context)s"):
            cls.context.enable(True)
            for handler in cls.handlers:
                handler.addFilter(cls.context.filter)

        else:
            for handler in cls.handlers:
                handler.removeFilter(cls.context.filter)

            cls.context.enable(False)

        if cls.is_using_format("%(pathname)s %(filename)s %(funcName)s %(module)s"):
            logging._srcfile = cls._logging_snapshot._srcfile

        else:
            logging._srcfile = None

        logging.logProcesses = cls.is_using_format("%(process)d")
        logging.logThreads = cls.is_using_format("%(thread)d %(threadName)s")

        if cls._shortcuts_fixed:
            return

        cls._shortcuts_fixed = True

        if hasattr(sys, "_getframe"):
            def log(level, msg, *args, **kwargs):
                """Wrapper to make logging.info() etc report the right module %(name)"""
                name = sys._getframe(1).f_globals.get("__name__")
                logger = logging.getLogger(name)
                try:
                    logging.currentframe = lambda: sys._getframe(4)
                    logger.log(level, msg, *args, **kwargs)

                finally:
                    logging.currentframe = ORIGINAL_CF

            def wrap(level, **kwargs):
                """Wrap corresponding logging shortcut function"""
                original = getattr(logging, logging.getLevelName(level).lower())
                f = partial(log, level, **kwargs)
                f.__doc__ = original.__doc__
                return f

            logging.critical = wrap(logging.CRITICAL)
            logging.fatal = logging.critical
            logging.error = wrap(logging.ERROR)
            logging.exception = partial(logging.error, exc_info=True)
            logging.warning = wrap(logging.WARNING)
            logging.info = wrap(logging.INFO)
            logging.debug = wrap(logging.DEBUG)
            logging.log = log


def _replace_and_pad(fmt, marker, replacement):
    """
    Args:
        fmt (str | unicode): Format to tweak
        marker (str | unicode): Marker to replace
        replacement (str | unicode): What to replace marker with

    Returns:
        (str): Resulting format, with marker replaced
    """
    if marker not in fmt:
        return fmt

    if replacement:
        return fmt.replace(marker, replacement)

    # Remove mention of 'marker' in 'fmt', including leading space (if present)
    fmt = fmt.replace("%s " % marker, marker)
    return fmt.replace(marker, "")


def _canonical_format(fmt):
    """
    Args:
        fmt (str | unicode | None): Format specification

    Returns:
        (str | None): Canonical version of format
    """
    if not fmt:
        return fmt

    return _replace_and_pad(fmt, "%(timezone)s", LogManager.spec.timezone)


def _get_file_handler(location, rotate, rotate_count):
    """
    Args:
        location (str | unicode | None): Log file path
        rotate (str | unicode | None): How to rotate, examples:
            time:midnight - Rotate at midnight
            time:15s - Rotate every 15 seconds
            time:2h - Rotate every 2 hours
            time:7d - Rotate every 7 days
            size:20m - Rotate every 20MB
            size:1g - Rotate every 1MB
        rotate_count (int): How many backups to keep

    Returns:
        (logging.FileHandler): Associated handler
    """
    if not rotate:
        return logging.FileHandler(location)

    kind, _, mode = rotate.partition(":")

    if not mode:
        raise ValueError("Invalid 'rotate' (missing kind): %s" % rotate)

    if kind == "time":
        if mode == "midnight":
            return TimedRotatingFileHandler(location, when="midnight", backupCount=rotate_count)

        timed = "shd"
        if mode[-1].lower() not in timed:
            raise ValueError("Invalid 'rotate' (unknown time spec): %s" % rotate)

        interval = to_int(mode[:-1])
        if interval is None:
            raise ValueError("Invalid 'rotate' (time range not an int): %s" % rotate)

        return TimedRotatingFileHandler(location, when=mode[-1], interval=interval, backupCount=rotate_count)

    if kind == "size":
        size = to_bytesize(mode)
        if size is None:
            raise ValueError("Invalid 'rotate' (size not a bytesize): %s" % rotate)

        return RotatingFileHandler(location, maxBytes=size, backupCount=rotate_count)

    raise ValueError("Invalid 'rotate' (unknown type): %s" % rotate)



"""
Import this only from your test cases

Example:

    from runez.conftest import cli, isolated_log_setup, temp_folder
"""

from __future__ import absolute_import

import logging
import os
import re
import sys
import tempfile

import _pytest.logging
import pytest

import runez

try:
    from click import BaseCommand as _ClickCommand
    from click.testing import CliRunner as _CliRunner

except ImportError:  # pragma: no cover, click used only if installed
    _ClickCommand = None
    _CliRunner = None


LOG = logging.getLogger(__name__)
TMP = tempfile.gettempdir()

# Set DEBUG logging level when running tests, makes sure LOG.debug() calls get captured (for inspection in tests)
logging.root.setLevel(logging.DEBUG)

if sys.argv and "pycharm" in sys.argv[0].lower():  # pragma: no cover
    # Ignore PyCharm's special wrapper "_jb_pytest_runner"...
    pt = runez.which("pytest")
    if pt:
        sys.argv[0] = pt

# Set logsetup defaults to stable/meaningful for pytest runs
runez.log.override_spec(timezone="UTC", tmp=TMP, locations=[os.path.join("{tmp}", "{basename}")])


class IsolatedLogSetup(object):
    """Allows to isolate changes to logging setup.

    This should only be useful for testing (as in general, logging setup is a global thing).
    """

    def __init__(self, adjust_tmp=True):
        self.adjust_tmp = adjust_tmp
        self.temp_folder = None

    def __enter__(self, tmp=False):
        """Context manager to save and restore log setup, useful for testing"""
        WrappedHandler.isolation += 1
        self.old_handlers = logging.root.handlers
        logging.root.handlers = []

        if self.adjust_tmp:
            # Adjust log.spec.tmp, and leave logging.root without any predefined handlers
            # intent: underlying wants to perform their own log.setup()
            self.old_spec = runez.log.spec.tmp
            self.temp_folder = runez.TempFolder()
            runez.log.spec.tmp = self.temp_folder.__enter__()
            return runez.log.spec.tmp

        # If we're not adjusting tmp, then make sure we have at least one logging handler defined
        handler = logging.StreamHandler(stream=sys.stderr)
        handler.setFormatter(logging.Formatter("%(levelname)s %(message)s"))
        handler.setLevel(logging.DEBUG)
        logging.root.addHandler(handler)

    def __exit__(self, *_):
        WrappedHandler.isolation -= 1
        runez.log.reset()
        logging.root.handlers = self.old_handlers

        if self.temp_folder:
            runez.log.spec.tmp = self.old_spec
            self.temp_folder.__exit__(*_)


@pytest.fixture
def cli():
    """Convenience fixture for click CLI testing.

    Example usage:

        from runez.conftest import cli
        from my_cli import main

        cli.default_main = main  # Handy if you have only one main

        def test_help(cli):
            cli.main = main  # Not needed if `cli.default_main` was set
            cli.run("--help")
            assert cli.succeeded
            assert cli.match("Usage:")

            # or more specifically
            assert "Usage:" in cli.logged.stdout
    """
    with cli.context() as context:
        yield ClickRunner(context=context)


# This just allows to get auto-complete to work in PyCharm
cli = cli  # type: ClickRunner

# Comes in handy for click apps with only one main entry point
cli.default_main = None

# Can be customized by users, wraps cli (fixture) runs in given context
cli.context = runez.TempFolder


@pytest.fixture
def isolated_log_setup():
    """Log settings restored"""
    with IsolatedLogSetup():
        yield


@pytest.fixture
def logged():
    with runez.CaptureOutput() as logged:
        yield logged


@pytest.fixture
def temp_folder():
    with runez.TempFolder() as tmp:
        yield tmp


# This just allows to get auto-complete to work in PyCharm
logged = logged  # type: runez.TrackedOutput
temp_folder = temp_folder  # type: str


class WrappedHandler(_pytest.logging.LogCaptureHandler):
    """pytest aggressively imposes its own capture, this allows to impose our capture where applicable"""

    isolation = 0

    def emit(self, record):
        if WrappedHandler.isolation == 0:
            if runez.context._CAPTURE_STACK:
                try:
                    msg = self.format(record)
                    stream = runez.context._CAPTURE_STACK[-1].buffer
                    stream.write(msg + "\n")
                    self.flush()
                except Exception:  # pragma: no cover
                    self.handleError(record)

            else:
                super(WrappedHandler, self).emit(record)

    @classmethod
    def find_wrapper(cls):
        """
        Returns:
            (WrappedHandler | None): WrappedHandler is logging.root, if present
        """
        if logging.root.handlers:
            for handler in logging.root.handlers:
                if handler and handler.__class__ is WrappedHandler:
                    return handler


_pytest.logging.LogCaptureHandler = WrappedHandler


class ClickWrapper(object):
    """Wrap click invoke, when click is available, otherwise just call provided function"""

    __runner = None

    def __init__(self, output=None, exit_code=None, exception=None):
        self.output = output
        self.exit_code = exit_code
        self.exception = exception

    def invoke(self, main, args):
        """Mocked click-like behavior"""
        try:
            try:
                output = main(*args)
                return ClickWrapper(output=output, exit_code=0)

            except TypeError as e:
                msg = runez.stringified(e)
                if msg and msg.startswith(main.__name__) and "takes" in msg and "arguments" in msg:
                    # py2: TypeError: hard_exit_no_args() takes no arguments (1 given)
                    # py3: TypeError: hard_exit_no_args() takes 0 positional arguments but 1 was given
                    old_argv = sys.argv
                    try:
                        sys.argv = old_argv[:1] + args
                        output = main()
                        return ClickWrapper(output=output, exit_code=0)

                    finally:
                        sys.argv = old_argv

                raise

        except SystemExit as e:
            if isinstance(e.code, int):
                exit_code = e.code
                output = None

            else:
                exit_code = 1
                output = runez.stringified(e)

            return ClickWrapper(output=output, exit_code=exit_code)

        except BaseException as e:
            return ClickWrapper(runez.stringified(e), exit_code=1, exception=e)

    @classmethod
    def new_runner(cls, main):
        """
        Returns:
            (ClickWrapper| click.testing.CliRunner): CliRunner if available
        """
        if _ClickCommand is not None and isinstance(main, _ClickCommand):
            return _CliRunner()

        return cls()


class ClickRunner(object):
    """Allows to provide a test-friendly fixture around testing click entry-points"""

    default_main = None  # This just allows to get auto-complete to work in PyCharm

    def __init__(self, context=None):
        """
        Args:
            context (callable | None): Context (example: temp folder) this click run was invoked under
        """
        self.context = context
        self.main = cli.default_main
        self.args = None  # type: list # Arguments used in last run() invocation
        self.logged = None  # type: runez.TrackedOutput
        self.exit_code = None  # type: int

    def run(self, *args, **kwargs):
        """
        Args:
            *args: Command line args
            **kwargs: If provided, format each arg with given `kwargs`
        """
        if kwargs:
            args = [runez.formatted(a, **kwargs) for a in args]

        if len(args) == 1 and hasattr(args[0], "split"):
            # Convenience: allow to provide full command as one string argument
            args = args[0].split()

        self.args = runez.flattened(args, split=runez.SHELL)

        with IsolatedLogSetup(adjust_tmp=False):
            with runez.CaptureOutput(dryrun=runez.DRYRUN) as logged:
                self.logged = logged
                if self.main is None:
                    self.main = self.default_main
                assert bool(self.main), "No main provided"
                runner = ClickWrapper.new_runner(self.main)
                result = runner.invoke(self.main, args=self.args)

                if result.output:
                    logged.stdout.buffer.write(result.output)

                if result.exception and not isinstance(result.exception, SystemExit):
                    try:
                        raise result.exception

                    except BaseException:
                        LOG.exception("Exited with stacktrace:")

                self.exit_code = result.exit_code

        if self.logged:
            handler = WrappedHandler.find_wrapper()
            if handler:
                handler.reset()
            title = runez.header("Captured output for: %s" % runez.represented_args(self.args), border="==")
            LOG.info("\n%s\nmain: %s\nexit_code: %s\n%s\n", title, self.main, self.exit_code, self.logged)

    @property
    def succeeded(self):
        return self.exit_code == 0

    @property
    def failed(self):
        return self.exit_code != 0

    def match(self, expected, stdout=None, stderr=None, regex=None):
        """
        Args:
            expected (str | unicode | re.Pattern): Message to find in self.logged
            stdout (bool | None): Look at stdout (default: yes, if captured)
            stderr (bool | None): Look at stderr (default: yes, if captured)
            regex (int | bool | None): Specify whether 'expected' should be a regex

        Returns:
            (Match | None): Match found, if any
        """
        if stdout is None and stderr is None:
            # By default, look at stdout/stderr only
            stdout = stderr = True

        assert expected, "No 'expected' provided"
        assert self.exit_code is not None, "run() was not called yet"

        captures = [stdout and self.logged.stdout, stderr and self.logged.stderr]
        captures = [c for c in captures if c is not None and c is not False]

        assert captures, "No captures specified"
        if not any(c for c in captures):
            # There was no output at all
            return None

        if not isinstance(regex, bool) and isinstance(regex, int):
            flags = regex
            regex = True

        else:
            flags = 0

        if isinstance(expected, runez.base.string_type) and "..." in expected and not isinstance(regex, bool):
            regex = True
            expected = expected.replace("...", ".+")

        if not isinstance(expected, runez.base.string_type):
            # Assume regex, no easy way to verify isinstance(expected, re.Pattern) for python < 3.7
            regex = expected

        elif regex:
            regex = re.compile("(.{0,32})(%s)(.{0,32})" % expected, flags=flags)

        for c in captures:
            contents = c.contents()
            if regex:
                m = regex.search(contents)
                if m:
                    if m.groups():
                        return Match(c, m.group(2), pre=m.group(1), post=m.group(3))

                    return Match(c, m.group(0))

            elif expected in contents:
                i = contents.index(expected)
                pre = runez.shortened(contents[:i], 32)
                post = runez.shortened(contents[i + len(expected):], 32)
                return Match(c, expected, pre=pre, post=post)

    def expect_messages(self, *expected, **kwargs):
        for message in expected:
            if message[0] == "!":
                m = self.match(message[1:], **kwargs)
                if m:
                    assert False, "Unexpected match in output: %s" % m

            else:
                m = self.match(message, **kwargs)
                if not m:
                    assert False, "Not seen in output: %s" % message

    def expect_success(self, args, *expected, **kwargs):
        spec = RunSpec()
        spec.pop(kwargs)
        self.run(args, **kwargs)
        assert self.succeeded, "%s failed, was expecting success" % runez.represented_args(self.args)
        self.expect_messages(*expected, **spec.to_dict())

    def expect_failure(self, args, *expected, **kwargs):
        spec = RunSpec()
        spec.pop(kwargs)
        self.run(args, **kwargs)
        assert self.failed, "%s succeeded, was expecting failure" % runez.represented_args(self.args)
        self.expect_messages(*expected, **spec.to_dict())


class RunSpec(runez.Slotted):

    _default = runez.UNSET

    __slots__ = ["stdout", "stderr", "regex"]


class Match(object):
    def __init__(self, capture, match, pre=None, post=None):
        self.capture = capture
        self.match = match
        self.pre = pre
        self.post = post

    def __repr__(self):
        return self.match
